import cv2
import numpy as np

from .utils import Utilities


class ImageExtractor:
    """This class is used to extract the image from the given black backgrounded image."""

    def __init__(self, image: np.ndarray) -> None:
        """Initializes the image extractor."""
        copied_image, half_width = self._read_image(image)
        cleared_bg = self._clear_background(copied_image)
        contour_mask = self._find_contour(cleared_bg, copied_image)
        min_x, max_x, min_y, max_y = self._calculate_edge_points(contour_mask)
        self._image = self._crop_image(
            image, min_x + half_width, max_x + half_width, min_y, max_y
        )

    def extract(self) -> np.ndarray:
        """Returns the extracted image."""
        return self._image

    @staticmethod
    def _read_image(image: np.ndarray) -> tuple[np.ndarray, int]:
        """Reads the image and returns it as a greyscale half right image."""
        # Convert to uin8.
        copied_image = Utilities.scale_back_from_float64_to_uint8(image.copy())
        # Convert the image to grayscale.
        if len(copied_image.shape) == 3:
            copied_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        # Crop the image only the right half of it and save its half width.
        half_width = copied_image.shape[1] // 2
        copied_image = copied_image[:, half_width:]
        return copied_image, half_width

    @staticmethod
    def _clear_background(image: np.ndarray) -> np.ndarray:
        """Clears the background of the image and returns it as a more cleaner backgrounded."""
        _, binary_mask = cv2.threshold(
            image, 0, 255, cv2.THRESH_OTSU + cv2.THRESH_BINARY_INV
        )
        # Erode the binary mask
        kernel = np.ones((15, 15), np.uint8)
        return cv2.morphologyEx(binary_mask, cv2.MORPH_OPEN, kernel, iterations=1)

    @staticmethod
    def _find_contour(mask: np.ndarray, gray_image: np.ndarray) -> np.ndarray:
        """Finds the contour of the image and returns it as a contour."""
        # Find contours in the eroded mask
        contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        # Sort the contours by area in descending order
        contours_sorted = sorted(contours, key=cv2.contourArea, reverse=True)
        # Find the biggest contour
        if len(contours_sorted) > 0:
            contour_to_show = contours_sorted[0]
            # Create a mask for the biggest contour
            contour_mask = np.bitwise_not(gray_image.copy())
            # Draw the biggest contour on the mask
            cv2.drawContours(contour_mask, [contour_to_show], 0, (0, 255, 0), -1)
            # Dilate the contour mask
            kernel = np.ones((100, 100), np.uint8)
            contour_mask = cv2.erode(contour_mask, kernel, iterations=1)
            contour_mask = cv2.dilate(contour_mask, kernel, iterations=1)
            # Apply binarization
            _, inv_contour_mask = cv2.threshold(
                contour_mask, 10, 255, cv2.THRESH_BINARY
            )

            return inv_contour_mask
        raise ValueError("No contour found.")

    @staticmethod
    def _calculate_edge_points(contour_mask: np.ndarray) -> tuple[int, int, int, int]:
        """Calculates the edge points of the image and returns it as a numpy array."""
        # Dilate one pixel, then substract the dilated image from the original mask to
        # obtain the inner boundary of the mask
        kernel = np.ones((1, 1), np.uint8)
        contour_mask_inner = cv2.dilate(contour_mask, kernel, iterations=1)
        contour_mask_inner = cv2.bitwise_not(contour_mask_inner)
        contour_mask_inner = cv2.subtract(contour_mask, contour_mask_inner)
        # Use Canny edge detection to find the edges of the inner boundary.
        edges = cv2.Canny(contour_mask_inner, 100, 200)
        # Find the coordinates of all the edge points
        edge_points = np.where(edges == 255)
        edge_points = np.array(edge_points).T

        # Find the minimum and maximum x and y coordinates
        min_x = np.min(edge_points[:, 1])
        max_x = np.max(edge_points[:, 1])
        min_y = np.min(edge_points[:, 0])
        max_y = np.max(edge_points[:, 0])
        return (min_x, max_x, min_y, max_y)

    @staticmethod
    def _crop_image(
        image: np.ndarray, min_x: int, max_x: int, min_y: int, max_y: int
    ) -> np.ndarray:
        """Crops the image and returns it as a cropped image."""
        return image[min_y:max_y, min_x:max_x]



"""
This module contains the Decoder class, which is responsible for decoding
secret images from one image.
"""
import numpy as np
from .utils import Utilities
from .decoder_base import RubiesDecoder


class SimpleDecoder(RubiesDecoder):
    """It handles the decoding functionality with a need of secret image sizes."""

    @staticmethod
    def _deinsert(whole_image, secret_image_sizes) -> np.ndarray:
        """It crops the secret images from the whole image."""
        v_pad = (whole_image.shape[0] - secret_image_sizes[0]) // 2
        h_pad = whole_image.shape[1] - secret_image_sizes[1]
        return whole_image[v_pad : v_pad + secret_image_sizes[0], h_pad - 1 : -1] // 100


"""
This module contains the BaseDecoder class, which is responsible for decoding
secret images from one image. Other decoders inherit from this class.
"""
import numpy as np
from .utils import Utilities


class RubiesDecoder:
    """It handles the decoding functionality without original image itself or secret image size."""

    def __init__(self, encoded_image_path: str) -> None:
        self._encoded_image = Utilities.read_image(encoded_image_path)

        # Get the LAB components of images.
        _, _enc_a, _enc_b = Utilities.split_to_lab_components(self._encoded_image)

        # Get the magnitude of chroma components.
        _enc_a_mag = Utilities.get_magnitude(_enc_a)
        _enc_b_mag = Utilities.get_magnitude(_enc_b)

        # Extract the difference.
        self._diff_a = self._extract_difference(_enc_a_mag)
        self._diff_b = self._extract_difference(_enc_b_mag)

        # Holds the decoded images.
        self._decoded_a: np.ndarray = None
        self._decoded_b: np.ndarray = None
    
    def decode(self, **kwargs) -> tuple[np.ndarray]:
        """It decodes the secret images from each chroma component."""
        # Crop the secret images.
        secret_image_a = self._deinsert(self._diff_a, **kwargs)
        secret_image_b = self._deinsert(self._diff_b, **kwargs)

        # Scale back the secret images.
        self._decoded_a = Utilities.scale_back_from_float64_to_uint8(secret_image_a)
        self._decoded_b = Utilities.scale_back_from_float64_to_uint8(secret_image_b)

        return (self._decoded_a, self._decoded_b)
    
    def save(self, path_a: str, path_b: str) -> None:
        """It saves the decoded images."""
        if self._decoded_a is None or self._decoded_b is None:
            raise ValueError("You should decode the images first.")
        Utilities.save_image(self._decoded_a, path_a)
        Utilities.save_image(self._decoded_b, path_b)

    @staticmethod
    def _extract_difference(modified_mag) -> np.ndarray:
        """It extracts the difference from the original and modified
        magnitudes."""
        average = np.mean(modified_mag)
        to_delete = 300
        modified_mag[0:to_delete, 0:to_delete] = average
        modified_mag[-to_delete:, -to_delete:] = average
        modified_mag[0:to_delete, -to_delete:] = average
        modified_mag[-to_delete:, 0:to_delete] = average
        modified_mag[modified_mag < average] = average
        return modified_mag - np.mean(modified_mag)
    
    @staticmethod
    def _deinsert() -> np.ndarray:
        """It crops the secret images from the whole image."""
        raise NotImplementedError("This method should be implemented in the child class.")



"""
This module contains the Decoder class, which is responsible for decoding
secret images from one image.
"""
import numpy as np
from .extractor import ImageExtractor
from .decoder_base import RubiesDecoder


class AutoDecoder(RubiesDecoder):
    """It handles the decoding functionality without original image itself or secret image size."""

    @staticmethod
    def _deinsert(whole_image) -> np.ndarray:
        """It crops the secret images from the whole image."""
        return ImageExtractor(whole_image).extract()



"""
Plotter class for plotting data.
"""
import numpy as np
import matplotlib.pyplot as plt
import cv2


def show_3d_magnitude_spectrum(magnitude_spectrum, title="Magnitude Spectrum"):
    """3D magnitude spectrum plotter."""
    fig = plt.figure(figsize=(8, 6))
    mag_plot = fig.add_subplot(111, projection="3d")

    x_vals = np.arange(0, magnitude_spectrum.shape[1], 1)
    y_vals = np.arange(0, magnitude_spectrum.shape[0], 1)
    x_mesh, y_mesh = np.meshgrid(x_vals, y_vals)

    db_spectrum = np.log10(1 + np.abs(magnitude_spectrum))

    mag_plot.plot_surface(x_mesh, y_mesh, db_spectrum, cmap="viridis")
    mag_plot.set_title(title)
    mag_plot.set_xlabel("X")
    mag_plot.set_ylabel("Y")
    mag_plot.set_zlabel("Magnitude (log(1+abs(mag)))")
    plt.show()


def show_difference_in_3d(
    original_version,
    modified_version,
    original_title="Original Magnitude Spectrum",
    modified_title="Magnitude Spectrum after IFFT",
):
    """Chrome-a and chrome-b magnitude spectrum plotter."""
    figure_window = plt.figure(figsize=(15, 6))

    original_plot = figure_window.add_subplot(121, projection="3d")
    x1_mesh, y1_mesh = np.meshgrid(
        np.arange(original_version.shape[1]), np.arange(original_version.shape[0])
    )
    original_plot.plot_surface(
        x1_mesh, y1_mesh, np.log(1 + np.abs(original_version)), cmap="viridis"
    )
    original_plot.set_title(original_title)
    original_plot.set_xlabel("X")
    original_plot.set_ylabel("Y")
    original_plot.set_zlabel("Magnitude (log(1+abs(mag)))")

    modified_plot = figure_window.add_subplot(122, projection="3d")
    x2_mesh, y2_mesh = np.meshgrid(
        np.arange(modified_version.shape[1]), np.arange(modified_version.shape[0])
    )
    modified_plot.plot_surface(
        x2_mesh, y2_mesh, np.log(1 + np.abs(modified_version)), cmap="viridis"
    )
    modified_plot.set_title(modified_title)
    modified_plot.set_xlabel("X")
    modified_plot.set_ylabel("Y")
    modified_plot.set_zlabel("Magnitude (log(1+abs(mag)))")

    plt.tight_layout()
    plt.show()


def show_image(title, image, ifft=False, phase=None):
    """Image shower."""
    plt.figure()
    plt.title(title)
    if ifft:
        image_to_show = np.abs(np.fft.ifft2(image * np.exp(1j * phase))).astype(
            np.uint8
        )
    else:
        image_to_show = image
    rgb_image = cv2.cvtColor(image_to_show, cv2.COLOR_BGR2RGB)
    plt.imshow(rgb_image)
    plt.axis("off")
    plt.show()



__all__ = [
    "Encoder",
    "SimpleDecoder",
    "AutoDecoder",
    "show_3d_magnitude_spectrum",
    "show_difference_in_3d",
    "show_image",
]

from .encoder import Encoder
from .decoder_simple import SimpleDecoder
from .decoder_auto import AutoDecoder
from .plotter import show_3d_magnitude_spectrum, show_difference_in_3d, show_image


"""
This module contains the Encoder class, which is responsible for encoding
images into one carrier image.
"""
import numpy as np
from .utils import Utilities


class Encoder:
    """It handles the encoding functionality."""

    def __init__(self, carrier_image: str | np.ndarray, secret_size: tuple) -> None:
        """It initializes the class."""
        self.secret_size = secret_size
        # Read the carrier image.
        self._image = Utilities.read_image(carrier_image)
        # Get the LAB components of the image.
        self._lum, _ch_a, _ch_b = Utilities.split_to_lab_components(self._image)
        self._ch_a_mag = Utilities.get_magnitude(_ch_a)
        self._ch_b_mag = Utilities.get_magnitude(_ch_b)
        self._ch_a_phase = Utilities.get_phase(_ch_a)
        self._ch_b_phase = Utilities.get_phase(_ch_b)
        # Store the encoded image.
        self._encoded_image: np.ndarray = None

    def encode(self, embed_img_a: str | np.ndarray, embed_img_b: str | np.ndarray) -> np.ndarray:
        """It encodes the images embedded into the carrier image."""
        # Recieve images.
        image_a = Utilities.read_image(embed_img_a, size=self.secret_size)
        image_b = Utilities.read_image(embed_img_b, size=self.secret_size)
        image_a = Utilities.rgb_to_gray(image_a)
        image_b = Utilities.rgb_to_gray(image_b)

        # Embed the images.
        _new_ch_a_mag = self._embed_onto("a", image_a)
        _new_ch_b_mag = self._embed_onto("b", image_b)

        # Create chroma components from the modified magnitude and phase.
        chrom_a = Utilities.ifft(_new_ch_a_mag, self._ch_a_phase)
        chrom_b = Utilities.ifft(_new_ch_b_mag, self._ch_b_phase)

        # Create the RGB image from LAB components.
        encoded_lab = Utilities.create_lab_from_complex(self._lum, chrom_a, chrom_b)
        self._encoded_image = Utilities.lab_to_rgb(encoded_lab)
        return self._encoded_image
    
    def save(self, path: str) -> None:
        """It saves the encoded image."""
        if self._encoded_image is None:
            raise ValueError("You should encode the images first.")
        Utilities.save_image(self._encoded_image, path)

    def _embed_onto(self, chroma_x: str, secret_image: np.ndarray) -> np.ndarray:
        """It embeds one image into carrier."""
        if chroma_x == "a":
            chroma_mag = self._ch_a_mag
        elif chroma_x == "b":
            chroma_mag = self._ch_b_mag
        else:
            raise ValueError("Invalid chroma component.")

        # Copy the magnitude spectum.
        new_chroma_mag = chroma_mag.copy()

        # Embed the secret image by adding on top of the freq magnitudes.
        return self._insert(new_chroma_mag, secret_image)

    @staticmethod
    def _insert(new_chroma_mags: np.ndarray, secret_image: np.ndarray) -> np.ndarray:
        """This is the logic where insertation has done."""
        # Calculate the padding.
        v_pad, h_pad = Utilities.calculate_padding(new_chroma_mags, secret_image)

        # Check if the secret image is too big to embed.
        if v_pad < 0 or h_pad < 0:
            raise ValueError("Secret image is too big to embed.")

        # Embed the secret image by adding on top of the freq magnitudes.
        secret_image = secret_image.astype(np.float64)
        secret_image *= 100
        # Add luminance to the secret image since
        # decoder will eliminate some.
        secret_image += 5000

        new_chroma_mags[
            v_pad : (v_pad + secret_image.shape[0]),
            h_pad - 1 : -1,
        ] += secret_image

        return new_chroma_mags



"""
This is a helper class that contains common functions for all the modules.
"""
import cv2
import numpy as np


class Utilities:
    """Common functions for all the modules."""

    @staticmethod
    def create_lab_from_complex(
        luminance: np.ndarray, chrom_a: np.ndarray, chrom_b: np.ndarray
    ) -> np.ndarray:
        """It creates the RGB image from LAB components."""
        real_chrom_a = np.abs(chrom_a).astype(np.uint8)
        real_chrom_b = np.abs(chrom_b).astype(np.uint8)
        return cv2.merge((luminance, real_chrom_a, real_chrom_b))

    @staticmethod
    def calculate_padding(max_elem: np.ndarray, min_elem: np.ndarray) -> list[int]:
        """It calculates the padding."""
        vertical_padding = (max_elem.shape[0] - min_elem.shape[0]) // 2
        horizontal_padding = max_elem.shape[1] - min_elem.shape[1]
        # vertical_padding = (max_elem.shape[0] - min_elem.shape[0]) // 2
        # horizontal_padding = (max_elem.shape[1] - min_elem.shape[1]) // 2
        return vertical_padding, horizontal_padding

    @staticmethod
    def split_to_lab_components(image: np.ndarray) -> list[np.ndarray]:
        """It splits the image to LAB components."""
        return cv2.split(Utilities.rgb_to_lab(image))

    @staticmethod
    def rgb_to_lab(rgb_image: np.ndarray) -> np.ndarray:
        """It converts the RGB image to LAB image."""
        return cv2.cvtColor(rgb_image, cv2.COLOR_BGR2LAB)

    @staticmethod
    def lab_to_rgb(lab_image: np.ndarray) -> np.ndarray:
        """It converts the LAB image to RGB image."""
        return cv2.cvtColor(lab_image, cv2.COLOR_LAB2BGR)

    @staticmethod
    def rgb_to_gray(rgb_image: np.ndarray) -> np.ndarray:
        """It converts the RGB image to gray image."""
        return cv2.cvtColor(rgb_image, cv2.COLOR_BGR2GRAY)

    @staticmethod
    def read_image(image: str | np.ndarray, size: tuple | None = None) -> np.ndarray:
        """It reads the image."""
        if isinstance(image, str):
            # Read the carrier image.
            return_image = cv2.imread(image)
            if return_image is None:
                raise FileNotFoundError("Image not found.")
        elif isinstance(image, np.ndarray):
            return_image = image
        else:
            raise TypeError("Invalid image type.")
        if size is not None:
            return_image = cv2.resize(return_image, size)
        return return_image
    
    @staticmethod
    def save_image(image: np.ndarray, path: str) -> None:
        """It saves the image."""
        # If the image path is .jpg, raise an error.
        if path.endswith(".jpg") or path.endswith(".jpeg"):
            raise ValueError("Please use loseless image type to save.")
        cv2.imwrite(path, image)

    @staticmethod
    def ifft(magnitudes: np.ndarray, phases: np.ndarray) -> np.ndarray:
        """It returns the IFFT of the image."""
        component = np.multiply(magnitudes, np.exp(1j * phases))
        return np.fft.ifft2(component)

    @staticmethod
    def fft(component: np.ndarray) -> np.ndarray:
        """It returns the FFT of the image."""
        return np.fft.fft2(component)

    @staticmethod
    def get_phase(fft_component: np.ndarray) -> np.ndarray:
        """It returns the phase of the FFT of the image."""
        return np.angle(Utilities.fft(fft_component))

    @staticmethod
    def get_magnitude(fft_component: np.ndarray) -> np.ndarray:
        """It returns the magnitude of the FFT of the image."""
        return np.abs(Utilities.fft(fft_component))

    @staticmethod
    def scale_back_from_float64_to_uint8(image: np.ndarray) -> np.ndarray:
        """It scales back the image from float64 to uint8."""
        # Scale the image np.ndarray from float64 to uint8 with using
        # peak to peak normalization.
        double_version = (255 * (image - np.min(image))) / np.ptp(image)
        return double_version.astype(np.uint8)
